#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "conntrack.h"
#include "newtable.h"

#if 0
#define printd(args...) printf(args)
#else
#define printd(args...)
#endif

/* Iterate over the buckets and their slots as we walk the tree
   Provides bucketcode and slotcode with various variables:
   	curtable
   	bucket
	bucketnr
	slotnr
   	hashshift
	hashbits
	depth
*/
#define hashtrie_iterate_hash(table, hash, bucketcode_enter, slotcode)		\
	do {									\
		struct hashentry *curtable = table, *bucket;			\
		int hashshift = 0;						\
		hashbits_t hashbits;						\
		u8 depth = 0;							\
		unsigned int slotnr, bucketnr;					\
										\
		while (curtable) {						\
			bucketnr = (hash >> hashshift) & (HASHNUM - 1);		\
			bucket = &curtable[bucketnr];				\
										\
			hashshift += HASHSHIFT;					\
			/* FIXME: This is wrong */				\
			if (hashshift > (sizeof(hash_t) * 8 - HASHBITS))	\
				hashshift -= sizeof(hash_t) * 8 - HASHBITS;	\
										\
			if (bucket->child)					\
				prefetchtest(&bucket->child[(hash >> hashshift)	\
					      & (HASHNUM - 1)]);		\
										\
			hashbits = (hash >> hashshift) & HASHBITMASK;		\
										\
			do {							\
				bucketcode_enter				\
			} while (0);						\
										\
			/* We want to start at entry 1 for depth 0		\
			   and entry 0 for all other depths */			\
			for (slotnr = !depth; slotnr < NUMENTRY; slotnr++) {	\
				do {						\
					slotcode				\
				} while (0);					\
			}							\
										\
			depth++;						\
			curtable = bucket->child;				\
		}								\
	} while (0)

#define hashtrie_iterate_all(table, bucketcode_leave, slotcode)			\
	do {									\
		struct hashentry *curtable = table, *bucket;			\
		struct hashentry_stack stack[MAXDEPTH];				\
		u8 depth = 0;							\
		unsigned int slotnr, bucketnr = 0;				\
										\
		for (;;) {							\
			stack[depth].bucket = curtable;				\
			stack[depth].bucketnr = bucketnr;			\
			printd("Current depth: %u table: %p bucketnr: %u\n", depth, curtable, bucketnr); \
										\
			bucket = &curtable[bucketnr];				\
										\
			if (bucket->child) {					\
				depth++;					\
				bucketnr = 0;					\
				curtable = bucket->child;			\
				printd("Following child to depth: %u table: %p bucketnr: %u\n", depth, curtable, bucketnr); \
				continue;					\
			}							\
										\
			for (slotnr = !depth; slotnr < NUMENTRY; slotnr++) {	\
				do {						\
					slotcode				\
				} while (0);					\
			}							\
										\
			bucketnr++;						\
			if (bucketnr >= HASHNUM) {				\
				if (depth == 0)					\
					break;					\
				depth--;					\
				curtable = stack[depth].bucket;			\
				bucketnr = stack[depth].bucketnr;		\
										\
				do {						\
					bucketcode_leave			\
				} while (0);					\
			}							\
		}								\
	} while (0)

struct ip_conntrack *lookup(struct hashtrie_info *info, struct ip_conntrack_tuple *tuple, hash_t hash)
{
	printd("\nEntering lookup()\n");
	printd("Looking up entry with hash: %x\n", hash);

	hashtrie_iterate_hash(info->hashtrie, hash, {}, {
		if (!bucket->data.members[slotnr])
			continue;

		printd("checking hashbits of %u\n", slotnr);
		printd("current hashbits: %x\n",
			bucket->hashbits[slotnr] & HASHBITMASK);
		
		if (hashbits == (bucket->hashbits[slotnr] & HASHBITMASK) &&
		    ip_ct_tuple_equal(tuple, bucket->data.members[slotnr])) {
			printd("Confirmed match %p at %u\n",
				bucket->data.members[slotnr], slotnr);
			return container_of(bucket->data.members[slotnr], struct ip_conntrack, tuple[bucket->data.members[slotnr]->dst.dir]);
		} else {
			printd("Miss at %u\n", slotnr);
		}
	});

	printd("Leaving lookup()\n\n\n");

	return NULL;
}

int alter(struct hashtrie_info *info, struct ip_conntrack_tuple *tuple, unsigned long status, hash_t hash)
{
	hashtrie_iterate_hash(info->hashtrie, hash, {}, {
		if (!bucket->data.members[slotnr])
			continue;

		printd("checking hashbits of %u\n", slotnr);
		printd("current hashbits: %x\n",
			bucket->hashbits[slotnr] & HASHBITMASK);
		
		if (hashbits == (bucket->hashbits[slotnr] & HASHBITMASK) &&
		    ip_ct_tuple_equal(tuple, bucket->data.members[slotnr])) {
			printd("Confirmed match %p at %u\n",
				bucket->data.members[slotnr], slotnr);
			bucket->hashbits[slotnr] = hashbits | ((status & 1) << HASHBITS);
			return 1;
		} else {
			printd("Miss at %u\n", slotnr);
		}
	});

	printd("Leaving alter()\n\n\n");

	return 0;
}


int evict(struct hashtrie_info *info, struct hashentry *bucket, u8 depth)
{
	unsigned int slotnr, evicted = 0;

	printd("\nEntering evict()\n");
	printd("Evicting from bucket %p at depth %u\n", bucket, depth);

	/* We want to start at entry 1 for depth 0 and entry 0 for all other depths */
	for (slotnr = !depth; slotnr < NUMENTRY; slotnr++) {
		if (!bucket->data.members[slotnr])
			continue;

		/* Try to find an entry to evict */
		if (!(bucket->hashbits[slotnr] & STATUSBITMASK)) {
			/* This is just to make newtest.c happy, should be ripped out */
			bucket->data.members[slotnr]->src.ip = 0;
			bucket->data.members[slotnr]->dst.ip = 0;

			delete(info, bucket->data.members[slotnr]);

			evicted++;
		}
	}

	printd("Leaving evict()\n\n\n");

	return evicted;
}

inline unsigned int expand(struct hashtrie_info *info, struct hashentry *bucket, uint8_t depth)
{
	struct hashentry *newtable;
	unsigned int ret;

	printd("\nEntering expand()\n");

	if (bucket->child) {
		printf("expand: Error, expanding a bucket with children?\n");
		return 0;
	}

	printd("expand: expanding bucket %p\n", bucket);
	
	ret = posix_memalign((void *)&newtable, CACHELINE_REAL, HASHNUM * sizeof(struct hashentry));
	if (ret != 0)
		return 0;

	memset(newtable, 0, HASHNUM * sizeof(struct hashentry));
	
	printd("expand: Newtable: %p\n", newtable);
	
	bucket->child = newtable;
	
	info->maxmem += HASHNUM * sizeof(struct hashentry);
	info->numchildren++;
	if (info->maxdepth < depth + 1)
		info->maxdepth = depth + 1;

	printd("Leaving expand()\n\n\n");

	return 1;
}

int insert(struct hashtrie_info *info, struct ip_conntrack_tuple *tuple, unsigned long status, hash_t hash, unsigned int force_evict)
{
	struct hashentry *insertbucket = NULL, *lockedbucket = NULL;
	struct hashentry_stack bucketstack[MAXDEPTH];
	unsigned int insertpos = 0, evicted = 0;
	int i;
	u8 maxdepth = 0;
	hashbits_t insertbits = 0;

	printd("\nEntering insert()\n");

	printd("Inserting entry %p with hash: %x\n", tuple, hash);

start:
	hashtrie_iterate_hash(info->hashtrie, hash, {
		bucketstack[depth].bucket = bucket;
		bucketstack[depth].bucketnr = bucketnr;

		maxdepth = depth;

		if (!lockedbucket) {
			lockedbucket = bucket;
			spin_lock(&(lockedbucket->data.lock));
		}
	}, {
		if (!bucket->data.members[slotnr]) {
			if (insertbucket == NULL) {
				insertbucket = bucket;
				insertpos = slotnr;
				insertbits = hashbits;
			}
			continue;
		}
		if (hashbits == (bucket->hashbits[slotnr] & HASHBITMASK) &&
		    ip_ct_tuple_equal(tuple, bucket->data.members[slotnr])) {
			printd("Error inserting duplicate entry, "
			       "original found at %u\n", slotnr);
			spin_unlock(&lockedbucket->data.lock);
			return 0;
		}
	});

	if (force_evict) {
		/* Walk backwards and evict the bucket if we havn't been able to evict anything
		   or if the bucket has the magic position in its hashtable, stop when we have
		   been able to evict something and the next bucket position isn't magic.
		   Should give somewhat fair distribution of evictions at the diffrent depths */
		for (i = maxdepth; i >= 0; i--) {
			if (!evicted || bucketstack[i].bucketnr == info->evictbucketnr) {
				evicted += evict(info, bucketstack[i].bucket, i);
				continue;
			}
			break;
		}

		if (!evicted) {
			/* We failed */
			printd("Eviction failed!\n");
			spin_unlock(&lockedbucket->data.lock);
			return 0;
		}
		force_evict = 0;
		insertbucket = NULL;
		goto start;		
	}

	/* Have we found a place to put our new entry? */
	if (insertbucket == NULL) {
		printd("Too many entries in bucket, needs expanding "
		       "(hashshift %u)\n", hashshift);
		if (!expand(info, bucketstack[maxdepth].bucket, maxdepth)) {
			spin_unlock(&lockedbucket->data.lock);
			return 0;
		}

		goto start;
	}

	insertbucket->hashbits[insertpos] = insertbits | ((status & 1) << HASHBITS);
	insertbucket->data.members[insertpos] = tuple;
	tuple->hashpointer = &insertbucket->data.members[insertpos];

	info->numentries++;

	printd("Inserting entry in bucket %p at %u with hashbits %x\n",
		insertbucket, insertpos, insertbits);

	spin_unlock(&lockedbucket->data.lock);

	printd("Leaving insert()\n\n\n");
	return 1;
}

inline void delete(struct hashtrie_info *info, struct ip_conntrack_tuple *tuple)
{
	*tuple->hashpointer = NULL;
	tuple->hashpointer = NULL;
	info->numentries--;
}

int allocate_hashtrie(struct hashtrie_info *info)
{
	unsigned int i, ret;

	ret = posix_memalign((void *)&info->hashtrie, CACHELINE_REAL, HASHNUM * sizeof(struct hashentry));
	if (ret != 0) {
		printf("Error allocating hashtrie\n");
		return 0;
	}

	memset(info->hashtrie, 0, HASHNUM * sizeof(struct hashentry));

	info->evictbucketnr = rand() % HASHNUM;

	for (i = 0; i < HASHNUM; i++)
		spin_lock_init(&info->hashtrie[i].data.lock);

	return 1;
}

int free_hashtrie(struct hashtrie_info *info)
{
	hashtrie_iterate_all(info->hashtrie, {
			/* Job for RCU */
			free(curtable[bucketnr].child);
			curtable[bucketnr].child = NULL;
			info->numchildren--;
	}, {
			if (!bucket->data.members[slotnr])
				continue;

			delete(info, bucket->data.members[slotnr]);
	});
				
	if (info->numentries || info->numchildren) {
		printf("Error unregistering hashtrie, not empty. entries: %u children %u\n", info->numentries, info->numchildren);
		return 0;
	}

	free(info->hashtrie);

	return 1;
}


