#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "rdtsc2.c"
#include "conntrack.h"
#include "newtable.h"

#define TESTNUM 100*1024
#define EVICTNUM 300*1024

int main(void)
{
	struct ip_conntrack **conntrack, **conntrack2, *retp;
	unsigned int ret, ret2, found = 0, notfound = 0;
	int i, dir, evicting;
	struct hashtrie_info info;

	rdtsc_cal(100000000);

	printf("sizeof struct hashentry: %u\n", sizeof(struct hashentry));
	printf("sizeof struct ip_conntrack: %u\n", sizeof(struct ip_conntrack));
	printf("number of conntrack entries: %u\n", TESTNUM);
	printf("number of slots per hashtrie bucket: %u\n", NUMENTRY);
	printf("number of pad bytes: %u\n", PADNUM);
	printf("number of bytes per child: %u\n", sizeof(struct hashentry) * HASHNUM);
	
	memset(&info, 0, sizeof(struct hashtrie_info));

	if (!allocate_hashtrie(&info)) {
		printf("Error registering hashtrie\n");
		return 1;
	}

	posix_memalign((void *)&conntrack, CACHELINE_REAL, sizeof(struct ip_conntrack *) * TESTNUM);
	//conntrack = (void *)malloc(sizeof(struct ip_conntrack *) * TESTNUM);
	if (!conntrack) {
		printf("Couldn't allocate conntracks\n");
		free_hashtrie(&info);
		return 1;
	}

	//memset(conntrack, 0, sizeof(struct ip_conntrack) * TESTNUM);

	posix_memalign((void *)&conntrack2, CACHELINE_REAL, sizeof(struct ip_conntrack *) * TESTNUM);
	//conntrack2 = (void *)malloc(sizeof(struct ip_conntrack *) * TESTNUM);
	if (!conntrack2) {
		printf("Couldn't allocate conntracks2\n");
		free_hashtrie(&info);
		return 1;
	}
	//memset(conntrack2, 0, sizeof(struct ip_conntrack) * TESTNUM);

	srand((int)time(NULL));

	ip_conntrack_hash_rnd = rand();

	for (i = 0; i < TESTNUM; i++) {
		//posix_memalign(&conntrack[i], CACHELINE_REAL, sizeof(struct ip_conntrack));
		conntrack[i] = malloc(sizeof(struct ip_conntrack));
		if (!conntrack[i]) {
			printf("Couldn't allocate conntrack\n");
			free_hashtrie(&info);
			return 1;
		}
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port = (u16)rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port = (u16)rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.protonum = 6;
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.dir = IP_CT_DIR_ORIGINAL;

		conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].src.u.tcp.port = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.u.tcp.port = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.protonum = 6;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.dir = IP_CT_DIR_REPLY;

		//posix_memalign(&conntrack2[i], CACHELINE_REAL, sizeof(struct ip_conntrack));
		conntrack2[i] = malloc(sizeof(struct ip_conntrack));
		if (!conntrack2[i]) {
			printf("Couldn't allocate conntrack2\n");
			free_hashtrie(&info);
			return 1;
		}
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port = (u16)rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port = (u16)rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.protonum = 6;
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.dir = IP_CT_DIR_ORIGINAL;

		conntrack2[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].src.u.tcp.port = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.ip = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.u.tcp.port = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.protonum = 6;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.dir = IP_CT_DIR_REPLY;
	}

	for (i = 0; i < TESTNUM; i++) {
		evicting = 0;
		if (info.numentries > EVICTNUM) /* two entries in hashtrie per conntrack */
			evicting = 1;
		TIMER(
			ret = insert(&info, &conntrack[i]->tuple[IP_CT_DIR_ORIGINAL], 0, hash_tuple(&conntrack[i]->tuple[IP_CT_DIR_ORIGINAL]), evicting);
			ret2 = insert(&info, &conntrack[i]->tuple[IP_CT_DIR_REPLY], 0, hash_tuple(&conntrack[i]->tuple[IP_CT_DIR_REPLY]), evicting);
		);
		if (!ret || !ret2) {
			//printf("Couldn't insert entry %p\n", &conntrack[i]);
			if (!ret) {
				conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = 0;
				info.insfail++;
			}
			if (!ret2) {
				conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip = 0;
				info.insfail++;
			}
			total_time -= code_timer;
			total_count--;
		}
	}
	
	printf("insert: ");
	print_timer();

	printf("Number of failed inserts: %u\n", info.insfail);
	printf("Number of entries in hashtrie: %u\n", info.numentries);
	printf("Number of children in hashtrie: %u\n", info.numchildren);
	printf("Maxdepth of hashtrie: %u (0 == root)\n", info.maxdepth);
	printf("Maximum memory usage: %u\n", info.maxmem);

	total_count = 0;
	total_time = 0;

	for (dir = 0; dir < IP_CT_DIR_MAX; dir++) {
		for (i = 0; i < TESTNUM; i++) {
			if (conntrack[i]->tuple[dir].src.ip && conntrack[i]->tuple[dir].dst.ip) {
				TIMER(ret = alter(&info, &conntrack[i]->tuple[dir], i % 2, hash_tuple(&conntrack[i]->tuple[dir])););
				if (!ret) {
					printf("Couldn't alter entry %p, dir: %u, i: %u\n", conntrack[i], dir, i);
					return 1;
				}
			}
		}
	}

	printf("alter: ");
	print_timer();

	total_count = 0;
	total_time = 0;

	for (dir = 0; dir < IP_CT_DIR_MAX; dir++) {
		for (i = 0; i < TESTNUM; i++) {
			if (conntrack[i]->tuple[dir].src.ip && conntrack[i]->tuple[dir].dst.ip) {
				TIMER(retp = lookup(&info, &conntrack[i]->tuple[dir], hash_tuple(&conntrack[i]->tuple[dir])););
				if (retp != conntrack[i]) {
					printf("Couldn't find entry %p, dir: %u, i: %u\n", conntrack[i], dir, i);
					return 1;
				}
			}
		}
	}

	printf("lookup: ");
	print_timer();

	total_count = 0;
	total_time = 0;

	for (dir = IP_CT_DIR_MAX - 1; dir >= 0; dir--) {
		for (i = TESTNUM - 1; i >= 0; i--) {
			if (conntrack[i]->tuple[dir].src.ip && conntrack[i]->tuple[dir].dst.ip) {
				TIMER(retp = lookup(&info, &conntrack[i]->tuple[dir], hash_tuple(&conntrack[i]->tuple[dir])););
				if (retp != conntrack[i]) {
					printf("Couldn't find entry %p, dir: %u, i: %u\n", conntrack[i], dir, i);
					return 1;
				}
			}
		}
	}

	printf("lookup (reverse): ");
	print_timer();

	total_count = 0;
	total_time = 0;

	for (dir = 0; dir < IP_CT_DIR_MAX; dir++) {
		for (i = 0; i < TESTNUM; i++) {
			if (conntrack2[i]->tuple[dir].src.ip && conntrack2[i]->tuple[dir].dst.ip) {
				TIMER(retp = lookup(&info, &conntrack2[i]->tuple[dir], hash_tuple(&conntrack2[i]->tuple[dir])););
				if (retp == conntrack2[i])
					found++;
				else
					notfound++;
			}
		}
	}

	printf("lookup (nomatch): ");
	print_timer();
	printf("Found entries in hashtrie: %d\n", found);
	printf("Not found entries in hashtrie: %d\n", notfound);

	total_count = 0;
	total_time = 0;

	for (i = 0; i < TESTNUM; i++) {
		if (conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip && conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip &&
		    conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip && conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip) {
			TIMER(
				delete(&info, &conntrack[i]->tuple[IP_CT_DIR_ORIGINAL]);
				delete(&info, &conntrack[i]->tuple[IP_CT_DIR_REPLY]);
			);
		}
	}

	printf("delete: ");
	print_timer();

	total_count = 0;
	total_time = 0;

	info.insfail = 0;

	for (i = 0; i < TESTNUM; i++) {
		evicting = 0;
		if (info.numentries > EVICTNUM) /* two entries in hashtrie per conntrack */
			evicting = 1;
		TIMER(
			ret = insert(&info, &conntrack[i]->tuple[IP_CT_DIR_ORIGINAL], 1, hash_tuple(&conntrack[i]->tuple[IP_CT_DIR_ORIGINAL]), evicting);
			ret2 = insert(&info, &conntrack[i]->tuple[IP_CT_DIR_REPLY], 1, hash_tuple(&conntrack[i]->tuple[IP_CT_DIR_REPLY]), evicting);
		);
		if (!ret || !ret2) {
			//printf("Couldn't insert entry %p\n", &conntrack[i]);
			if (!ret) {
				conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = 0;
				info.insfail++;
			}
			if (!ret2) {
				conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip = 0;
				info.insfail++;
			}
			total_time -= code_timer;
			total_count--;
			info.insfail++;
		}
	}
	
	printf("insert2: ");
	print_timer();

	printf("Number of failed inserts: %u\n", info.insfail);
	printf("Number of entries in hashtrie: %u\n", info.numentries);
	printf("Number of children in hashtrie: %u\n", info.numchildren);
	printf("Maxdepth of hashtrie: %u (0 == root)\n", info.maxdepth);
	printf("Maximum memory usage: %u\n", info.maxmem);

	if (!free_hashtrie(&info)) {
		return 1;
	}

	for (i = 0; i < TESTNUM; i++) {
		free(conntrack[i]);
		free(conntrack2[i]);
	}
	free(conntrack);
	free(conntrack2);

	return 0;
}
