#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "rdtsc2.c"
#include "conntrack.h"
#include "oldtable.h"

/* 16 times the number of buckets is the normal ratio for a fully loaded untuned system */
#define TESTNUM BUCKETS

int main(void)
{
	struct ip_conntrack **conntrack, **conntrack2, *retp;
	unsigned int ret, ret2, found = 0, notfound = 0;
	int i, dir;
	struct hashtable_info info;

	rdtsc_cal(100000000);

	printf("sizeof struct ip_conntrack: %u\n", sizeof(struct ip_conntrack));
	printf("number of buckets in hashtable: %u\n", BUCKETS);
	printf("number of conntrack entries: %u\n", TESTNUM);
	
	memset(&info, 0, sizeof(struct hashtable_info));

	if (!register_hashtable(&info)) {
		printf("Error registering hashtable\n");
		return 1;
	}

	posix_memalign((void *)&conntrack, CACHELINE_REAL, sizeof(struct ip_conntrack *) * TESTNUM);
	//conntrack = (void *)malloc(sizeof(struct ip_conntrack) * TESTNUM);
	if (!conntrack) {
		printf("Couldn't allocate conntracks\n");
		unregister_hashtable(&info);
		return 1;
	}

	posix_memalign((void *)&conntrack2, CACHELINE_REAL, sizeof(struct ip_conntrack *) * TESTNUM);
	//conntrack2 = (void *)malloc(sizeof(struct ip_conntrack) * TESTNUM);
	if (!conntrack2) {
		printf("Couldn't allocate conntracks2\n");
		unregister_hashtable(&info);
		return 1;
	}

	srand((int)time(NULL));

	ip_conntrack_hash_rnd = rand();

	for (i = 0; i < TESTNUM; i++) {
		//posix_memalign((void *)&conntrack[i], CACHELINE_REAL, sizeof(struct ip_conntrack));
		conntrack[i] = malloc(sizeof(struct ip_conntrack));
		if (!conntrack[i]) {
			printf("Couldn't allocate conntrack\n");
			unregister_hashtable(&info);
			return 1;
		}

		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port = (u16)rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port = (u16)rand();
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.protonum = 6;
		conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.dir = IP_CT_DIR_ORIGINAL;
		INIT_LIST_HEAD(&conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].list);

		conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].src.u.tcp.port = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.u.tcp.port = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.protonum = 6;
		conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.dir = IP_CT_DIR_REPLY;
		INIT_LIST_HEAD(&conntrack[i]->tuple[IP_CT_DIR_REPLY].list);

		//posix_memalign((void *)&conntrack2[i], CACHELINE_REAL, sizeof(struct ip_conntrack));
		conntrack2[i] = malloc(sizeof(struct ip_conntrack));
		if (!conntrack2[i]) {
			printf("Couldn't allocate conntrack2\n");
			unregister_hashtable(&info);
			return 1;
		}

		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port = (u16)rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port = (u16)rand();
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.protonum = 6;
		conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.dir = IP_CT_DIR_ORIGINAL;
		INIT_LIST_HEAD(&conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].list);

		conntrack2[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].src.u.tcp.port = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].dst.u.tcp.port;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.ip = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.u.tcp.port = conntrack2[i]->tuple[IP_CT_DIR_ORIGINAL].src.u.tcp.port;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.protonum = 6;
		conntrack2[i]->tuple[IP_CT_DIR_REPLY].dst.dir = IP_CT_DIR_REPLY;
		INIT_LIST_HEAD(&conntrack2[i]->tuple[IP_CT_DIR_REPLY].list);
	}

	for (i = 0; i < TESTNUM; i++) {
		TIMER(
			ret = insert(&info, &conntrack[i]->tuple[IP_CT_DIR_ORIGINAL], hash_tuple(&conntrack[i]->tuple[IP_CT_DIR_ORIGINAL]));
			ret2 = insert(&info, &conntrack[i]->tuple[IP_CT_DIR_REPLY], hash_tuple(&conntrack[i]->tuple[IP_CT_DIR_REPLY]));
		);
		if (!ret || !ret2) {
			//printf("Couldn't insert entry %p\n", conntrack[i]);
			if (!ret)
				conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip = conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip = 0;
			if (!ret2)
				conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip = conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip = 0;
			total_time -= code_timer;
			total_count--;
		}
	}
	
	printf("insert: ");
	print_timer();

	printf("Number of entries in hashtable: %u\n", info.numentries);
	printf("Longest search in hashtable: %u\n", info.longsearch);
	info.longsearch = 0;

	total_count = 0;
	total_time = 0;

	for (dir = 0; dir < IP_CT_DIR_MAX; dir++) {
		for (i = 0; i < TESTNUM; i++) {
			if (conntrack[i]->tuple[dir].src.ip && conntrack[i]->tuple[dir].dst.ip) {
				TIMER(retp = lookup(&info, &conntrack[i]->tuple[dir], hash_tuple(&conntrack[i]->tuple[dir])););
				if (retp != conntrack[i]) {
					printf("Couldn't find entry %p\n", conntrack[i]);
					return 1;
				}
			}
		}
	}

	printf("lookup: ");
	print_timer();
	printf("Longest search in hashtable: %u\n", info.longsearch);
	info.longsearch = 0;

	total_count = 0;
	total_time = 0;

	for (dir = IP_CT_DIR_MAX - 1; dir >= 0; dir--) {
		for (i = TESTNUM - 1; i >= 0; i--) {
			if (conntrack[i]->tuple[dir].src.ip && conntrack[i]->tuple[dir].dst.ip) {
				TIMER(retp = lookup(&info, &conntrack[i]->tuple[dir], hash_tuple(&conntrack[i]->tuple[dir])););
				if (retp != conntrack[i]) {
					printf("Couldn't find entry %p\n", conntrack[i]);
					return 1;
				}
			}
		}
	}

	printf("lookup (reverse): ");
	print_timer();
	printf("Longest search in hashtable: %u\n", info.longsearch);
	info.longsearch = 0;

	total_count = 0;
	total_time = 0;

	for (dir = 0; dir < IP_CT_DIR_MAX; dir++) {
		for (i = 0; i < TESTNUM; i++) {
			if (conntrack2[i]->tuple[dir].src.ip && conntrack2[i]->tuple[dir].dst.ip) {
				TIMER(retp = lookup(&info, &conntrack2[i]->tuple[dir], hash_tuple(&conntrack2[i]->tuple[dir])););
				if (retp == conntrack2[i])
					found++;
				else
					notfound++;
			}
		}
	}

	printf("lookup (nomatch): ");
	print_timer();
	printf("Found entries in hashtable: %d\n", found);
	printf("Not found entries in hashtable: %d\n", notfound);

	total_count = 0;
	total_time = 0;

	for (i = 0; i < TESTNUM; i++) {
		if (conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].src.ip && conntrack[i]->tuple[IP_CT_DIR_ORIGINAL].dst.ip &&
		    conntrack[i]->tuple[IP_CT_DIR_REPLY].src.ip && conntrack[i]->tuple[IP_CT_DIR_REPLY].dst.ip) {
			TIMER(
				ret = delete(&info, &conntrack[i]->tuple[IP_CT_DIR_ORIGINAL]);
				ret2 = delete(&info, &conntrack[i]->tuple[IP_CT_DIR_REPLY]);
			);
			if (!ret || !ret2) {
				printf("Couldn't delete entry %p (entries: %u)\n", conntrack[i], info.numentries);
				return 1;
			}
		}
	}

	printf("delete: ");
	print_timer();

	printf("Number of entries in hashtable: %u\n", info.numentries);

	if (!unregister_hashtable(&info)) {
		printf("Error unregistering hashtable\n");
		return 1;
	}
	free(conntrack);
	free(conntrack2);

	return 0;
}
