#include <math.h>
#include <sys/time.h>
#include <time.h>

static unsigned long total_count;
static unsigned long long total_time;
static unsigned long long code_timer;

#define TIMER(code) \
	do { \
		code_timer = rdtsc_read(); \
		do{code}while(0); \
		code_timer = rdtsc_read() - code_timer; \
		total_time += code_timer; \
		total_count++; \
	} while (0)


static double ns_per_clock = 0.0;

/*
 * Do not depend on kernel header. Taken from 2.4.19-pre7 include/asm/msr.h
 */
#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
     
#define rdtscll(val) \
     __asm__ __volatile__("rdtsc" : "=A" (val))

static inline unsigned long long rdtsc_read(void)
{
	unsigned long high, low;
	rdtsc(low,high);
	return ((unsigned long long)high << 32) | low;
}

static inline unsigned long long rdtsc_read2(void)
{
	unsigned long long val;
	rdtscll(val);
	return val;
}

static unsigned long long rdtsc__cal_empty(int cnt)
{
	unsigned long long start = rdtsc_read();
	while (cnt-- > 0) ;
	return rdtsc_read() - start;
}

static int rdtsc__cal_run(int cnt)
{
	struct timeval tv_start;
	struct timeval tv_delta;
	unsigned long long delta;
	unsigned long long delta_ns;
	double prev_ns_per_clock = ns_per_clock;

	gettimeofday(&tv_start, 0);
	delta = rdtsc__cal_empty(cnt);
	gettimeofday(&tv_delta, 0);
	tv_delta.tv_sec -= tv_start.tv_sec;
	if (tv_start.tv_usec <= tv_delta.tv_usec) {
		tv_delta.tv_usec -= tv_start.tv_usec;
	} else {
		tv_delta.tv_usec += (1000000 - tv_start.tv_usec);
		tv_delta.tv_sec--;
	}
	delta_ns = 1000000000ull * tv_delta.tv_sec;
	delta_ns += 1000ull * tv_delta.tv_usec;
	ns_per_clock = (1.0 * delta_ns) / delta;
#ifdef RDTSC_VERBOSE
	printf("rdtsc_cal(%d), ns/clock old=%.2f new=%.2f, %.0fmhz, waste %.3fus\n",
		cnt,
		prev_ns_per_clock,
		ns_per_clock,
		1000 / ns_per_clock,
		cnt * ns_per_clock / 1000.0);
#endif
	return (fabs(ns_per_clock - prev_ns_per_clock) < (0.0005 * ns_per_clock));
}

void rdtsc_cal(int cntmax)
{
	int cnt = 1 << 10;
	unsigned long long cyc;
	struct timeval tv_start;
	struct timeval tv_delta;

	gettimeofday(&tv_start, 0);
	cyc = rdtsc_read();
	while (cnt < cntmax) {
		if (rdtsc__cal_run(cnt))
			break;
		cnt <<= 1;
	}
	cyc = rdtsc_read() - cyc;
	gettimeofday(&tv_delta, 0);
	tv_delta.tv_sec -= tv_start.tv_sec;
	if (tv_start.tv_usec <= tv_delta.tv_usec) {
		tv_delta.tv_usec -= tv_start.tv_usec;
	} else {
		tv_delta.tv_usec += (1000000 - tv_start.tv_usec);
		tv_delta.tv_sec--;
	}
#ifdef RDTSC_VERBOSE
	printf("rdtsc_cal: %.2f ns/cyc, took %Lu cyc / %.3f ms\n",
		ns_per_clock,
		cyc,
		cyc * ns_per_clock / 1000000.0);
	printf("rdtsc_cal gettimeofday interval: %.3f ms\n",
		(tv_delta.tv_sec * 1000000.0 + tv_delta.tv_usec) / 1000.0);
#endif
}

void print_timer()
{               
	printf("time: %Lu cyc, %.0f ns (%.0f/s)\n",
		total_time / total_count,
		(ns_per_clock * total_time) / total_count,
		1e9 / (ns_per_clock * total_time / total_count));
};      

